@@ -1,6 +1,6 @@
 /*
- * Copyright © 2010-2014 Inria.  All rights reserved.
- * Copyright © 2011-2012 Université Bordeaux 1
+ * Copyright © 2010-2017 Inria.  All rights reserved.
+ * Copyright © 2011-2012 Université Bordeaux
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -9,6 +9,7 @@
 #include <hwloc.h>
 #include <private/private.h>
 #include <private/debug.h>
+#include <private/misc.h>
 
 #include <float.h>
 #include <math.h>
@@ -84,6 +85,7 @@ void hwloc_distances_set(hwloc_topology_t __hwloc_restrict topology, hwloc_obj_t
   if (!nbobjs)
     /* we're just clearing, return now */
     return;
+  assert(nbobjs >= 2);
 
   /* create the new element */
   osdist = malloc(sizeof(struct hwloc_os_distances_s));
@@ -119,12 +121,12 @@ static int hwloc_distances__check_matrix(hwloc_topology_t __hwloc_restrict topol
 }
 
 static void hwloc_distances__set_from_string(struct hwloc_topology *topology,
-					     hwloc_obj_type_t type, char *string)
+					     hwloc_obj_type_t type, const char *string)
 {
   /* the string format is: "index[0],...,index[N-1]:distance[0],...,distance[N*N-1]"
    * or "index[0],...,index[N-1]:X*Y" or "index[0],...,index[N-1]:X*Y*Z"
    */
-  char *tmp = string, *next;
+  const char *tmp = string, *next;
   unsigned *indexes;
   float *distances;
   unsigned nbobjs = 0, i, j, x, y, z;
@@ -136,7 +138,21 @@ static void hwloc_distances__set_from_string(struct hwloc_topology *topology,
 
   if (sscanf(string, "%u-%u:", &i, &j) == 2) {
     /* range i-j */
+    if (j <= i) {
+      fprintf(stderr, "Ignoring %s distances from environment variable, range doesn't cover at least 2 indexes\n",
+	      hwloc_obj_type_string(type));
+      return;
+    }
     nbobjs = j-i+1;
+
+    tmp = strchr(string, ':');
+    if (!tmp) {
+      fprintf(stderr, "Ignoring %s distances from environment variable, missing colon\n",
+	      hwloc_obj_type_string(type));
+      return;
+    }
+    tmp++;
+
     indexes = calloc(nbobjs, sizeof(unsigned));
     distances = calloc(nbobjs*nbobjs, sizeof(float));
     /* make sure the user didn't give a veeeeery large range */
@@ -147,12 +163,14 @@ static void hwloc_distances__set_from_string(struct hwloc_topology *topology,
     }
     for(j=0; j<nbobjs; j++)
       indexes[j] = j+i;
-    tmp = strchr(string, ':') + 1;
 
   } else {
     /* explicit list of indexes, count them */
     while (1) {
       size_t size = strspn(tmp, "0123456789");
+      if (!size)
+	/* missing index */
+	break;
       if (tmp[size] != ',') {
 	/* last element */
 	tmp += size;
@@ -164,6 +182,11 @@ static void hwloc_distances__set_from_string(struct hwloc_topology *topology,
       nbobjs++;
     }
 
+    if (nbobjs < 2) {
+      fprintf(stderr, "Ignoring %s distances from environment variable, needs at least 2 indexes\n",
+	      hwloc_obj_type_string(type));
+      return;
+    }
     if (*tmp != ':') {
       fprintf(stderr, "Ignoring %s distances from environment variable, missing colon\n",
 	      hwloc_obj_type_string(type));
@@ -176,7 +199,7 @@ static void hwloc_distances__set_from_string(struct hwloc_topology *topology,
 
     /* parse indexes */
     for(i=0; i<nbobjs; i++) {
-      indexes[i] = strtoul(tmp, &next, 0);
+      indexes[i] = strtoul(tmp, (char **) &next, 0);
       tmp = next+1;
     }
   }
@@ -238,7 +261,8 @@ void hwloc_distances_set_from_env(struct hwloc_topology *topology)
 {
   hwloc_obj_type_t type;
   for(type = HWLOC_OBJ_SYSTEM; type < HWLOC_OBJ_TYPE_MAX; type++) {
-    char *env, envname[64];
+    const char *env;
+    char envname[64];
     snprintf(envname, sizeof(envname), "HWLOC_%s_DISTANCES", hwloc_obj_type_string(type));
     env = getenv(envname);
     if (env) {
@@ -266,7 +290,7 @@ int hwloc_topology_set_distance_matrix(hwloc_topology_t __hwloc_restrict topolog
     return 0;
   }
 
-  if (!nbobjs || !indexes || !distances)
+  if (nbobjs < 2 || !indexes || !distances)
     return -1;
 
   if (hwloc_distances__check_matrix(topology, type, nbobjs, indexes, NULL, distances) < 0)
@@ -449,73 +473,61 @@ void hwloc_distances_finalize_os(struct hwloc_topology *topology)
  * into exported logical distances attached to objects
  */
 
-static hwloc_obj_t
-hwloc_get_obj_covering_cpuset_nodeset(struct hwloc_topology *topology,
-				      hwloc_const_cpuset_t cpuset,
-				      hwloc_const_nodeset_t nodeset)
-{
-  hwloc_obj_t parent = hwloc_get_root_obj(topology), child;
-
-  assert(cpuset);
-  assert(nodeset);
-  assert(hwloc_bitmap_isincluded(cpuset, parent->cpuset));
-  assert(!nodeset || hwloc_bitmap_isincluded(nodeset, parent->nodeset));
-
- trychildren:
-  child = parent->first_child;
-  while (child) {
-    /* look for a child with a cpuset containing ours.
-     * if it has a nodeset, it must also contain ours.
-     */
-    if (child->cpuset && hwloc_bitmap_isincluded(cpuset, child->cpuset)
-	&& (!child->nodeset || hwloc_bitmap_isincluded(nodeset, child->nodeset))) {
-      parent = child;
-      goto trychildren;
-    }
-    child = child->next_sibling;
-  }
-  return parent;
-}
-
 static void
 hwloc_distances__finalize_logical(struct hwloc_topology *topology,
 				  unsigned nbobjs,
 				  hwloc_obj_t *objs, float *osmatrix)
 {
+  struct hwloc_distances_s ** tmpdistances;
   unsigned i, j, li, lj, minl;
   float min = FLT_MAX, max = FLT_MIN;
-  hwloc_obj_t root;
+  hwloc_obj_t root, obj;
   float *matrix;
-  hwloc_cpuset_t cpuset;
-  hwloc_nodeset_t nodeset;
-  unsigned relative_depth;
+  hwloc_cpuset_t cpuset, complete_cpuset;
+  hwloc_nodeset_t nodeset, complete_nodeset;
+  unsigned depth;
   int idx;
 
   /* find the root */
   cpuset = hwloc_bitmap_alloc();
+  complete_cpuset = hwloc_bitmap_alloc();
   nodeset = hwloc_bitmap_alloc();
+  complete_nodeset = hwloc_bitmap_alloc();
   for(i=0; i<nbobjs; i++) {
     hwloc_bitmap_or(cpuset, cpuset, objs[i]->cpuset);
+    if (objs[i]->complete_cpuset)
+      hwloc_bitmap_or(complete_cpuset, complete_cpuset, objs[i]->complete_cpuset);
     if (objs[i]->nodeset)
       hwloc_bitmap_or(nodeset, nodeset, objs[i]->nodeset);
+    if (objs[i]->complete_nodeset)
+      hwloc_bitmap_or(complete_nodeset, complete_nodeset, objs[i]->complete_nodeset);
   }
-  /* find the object covering cpuset AND nodeset (can't use hwloc_get_obj_covering_cpuset()) */
-  root = hwloc_get_obj_covering_cpuset_nodeset(topology, cpuset, nodeset);
+  /* find the object covering cpuset, we'll take care of the nodeset later */
+  root = hwloc_get_obj_covering_cpuset(topology, cpuset);
+  /* walk up to find a parent that also covers the nodeset */
+  while (root &&
+	 (!hwloc_bitmap_isincluded(nodeset, root->nodeset)
+	  || !hwloc_bitmap_isincluded(complete_nodeset, root->complete_nodeset)
+	  || !hwloc_bitmap_isincluded(complete_cpuset, root->complete_cpuset)))
+    root = root->parent;
   if (!root) {
-    /* should not happen, ignore the distance matrix and report an error. */
-    if (!hwloc_hide_errors()) {
+    /* should not happen unless cpuset is empty.
+     * cpuset may be empty if some objects got removed by KEEP_STRUCTURE, etc
+     * or if all objects are CPU-less.
+     */
+    if (!hwloc_hide_errors() && !hwloc_bitmap_iszero(cpuset)) {
       char *a, *b;
       hwloc_bitmap_asprintf(&a, cpuset);
       hwloc_bitmap_asprintf(&b, nodeset);
       fprintf(stderr, "****************************************************************************\n");
-      fprintf(stderr, "* hwloc has encountered an error when adding a distance matrix to the topology.\n");
+      fprintf(stderr, "* hwloc %s has encountered an error when adding a distance matrix to the topology.\n", HWLOC_VERSION);
       fprintf(stderr, "*\n");
       fprintf(stderr, "* hwloc_distances__finalize_logical() could not find any object covering\n");
       fprintf(stderr, "* cpuset %s and nodeset %s\n", a, b);
       fprintf(stderr, "*\n");
       fprintf(stderr, "* Please report this error message to the hwloc user's mailing list,\n");
 #ifdef HWLOC_LINUX_SYS
-      fprintf(stderr, "* along with the output from the hwloc-gather-topology script.\n");
+      fprintf(stderr, "* along with the files generated by the hwloc-gather-topology script.\n");
 #else
       fprintf(stderr, "* along with any relevant topology information from your platform.\n");
 #endif
@@ -524,7 +536,9 @@ hwloc_distances__finalize_logical(struct hwloc_topology *topology,
       free(b);
     }
     hwloc_bitmap_free(cpuset);
+    hwloc_bitmap_free(complete_cpuset);
     hwloc_bitmap_free(nodeset);
+    hwloc_bitmap_free(complete_nodeset);
     return;
   }
   /* don't attach to Misc objects */
@@ -533,16 +547,32 @@ hwloc_distances__finalize_logical(struct hwloc_topology *topology,
   /* ideally, root has the exact cpuset and nodeset.
    * but ignoring or other things that remove objects may cause the object array to reduce */
   assert(hwloc_bitmap_isincluded(cpuset, root->cpuset));
+  assert(hwloc_bitmap_isincluded(complete_cpuset, root->complete_cpuset));
   assert(hwloc_bitmap_isincluded(nodeset, root->nodeset));
+  assert(hwloc_bitmap_isincluded(complete_nodeset, root->complete_nodeset));
   hwloc_bitmap_free(cpuset);
+  hwloc_bitmap_free(complete_cpuset);
   hwloc_bitmap_free(nodeset);
-  if (root->depth >= objs[0]->depth) {
+  hwloc_bitmap_free(complete_nodeset);
+  depth = objs[0]->depth; /* this assume that we have distances between objects of the same level */
+  if (root->depth >= depth) {
     /* strange topology led us to find invalid relative depth, ignore */
     return;
   }
-  relative_depth = objs[0]->depth - root->depth; /* this assume that we have distances between objects of the same level */
 
-  if (nbobjs != hwloc_get_nbobjs_inside_cpuset_by_depth(topology, root->cpuset, root->depth + relative_depth))
+  /* count objects at that depth that are below root.
+   * we can't use hwloc_get_nbobjs_inside_cpuset_by_depth() because it ignore CPU-less objects.
+   */
+  i = 0;
+  obj = NULL;
+  while ((obj = hwloc_get_next_obj_by_depth(topology, depth, obj)) != NULL) {
+    hwloc_obj_t myparent = obj->parent;
+    while (myparent->depth > root->depth)
+      myparent = myparent->parent;
+    if (myparent == root)
+      i++;
+  }
+  if (i != nbobjs)
     /* the root does not cover the right number of objects, maybe we failed to insert a root (bad intersect or so). */
     return;
 
@@ -571,10 +601,14 @@ hwloc_distances__finalize_logical(struct hwloc_topology *topology,
   }
 
   /* store the normalized latency matrix in the root object */
+  tmpdistances = realloc(root->distances, (root->distances_count+1) * sizeof(struct hwloc_distances_s *));
+  if (!tmpdistances)
+    return; /* Failed to allocate, ignore this distance matrix */
+
+  root->distances = tmpdistances;
   idx = root->distances_count++;
-  root->distances = realloc(root->distances, root->distances_count * sizeof(struct hwloc_distances_s *));
   root->distances[idx] = malloc(sizeof(struct hwloc_distances_s));
-  root->distances[idx]->relative_depth = relative_depth;
+  root->distances[idx]->relative_depth = depth - root->depth;
   root->distances[idx]->nbobjs = nbobjs;
   root->distances[idx]->latency = matrix = malloc(nbobjs*nbobjs*sizeof(float));
   root->distances[idx]->latency_base = (float) min;
@@ -653,7 +687,7 @@ static void hwloc_report_user_distance_error(const char *msg, int line)
 
     if (!reported && !hwloc_hide_errors()) {
         fprintf(stderr, "****************************************************************************\n");
-        fprintf(stderr, "* hwloc has encountered what looks like an error from user-given distances.\n");
+        fprintf(stderr, "* hwloc %s has encountered what looks like an error from user-given distances.\n", HWLOC_VERSION);
         fprintf(stderr, "*\n");
         fprintf(stderr, "* %s\n", msg);
         fprintf(stderr, "* Error occurred in topology.c line %d\n", line);
@@ -832,6 +866,7 @@ hwloc__groups_by_distances(struct hwloc_topology *topology,
       hwloc_obj_t *groupobjs = NULL;
       unsigned *groupsizes = NULL;
       float *groupdistances = NULL;
+      unsigned failed = 0;
 
       groupobjs = malloc(sizeof(hwloc_obj_t) * nbgroups);
       groupsizes = malloc(sizeof(unsigned) * nbgroups);
@@ -851,22 +886,39 @@ hwloc__groups_by_distances(struct hwloc_topology *topology,
 	    if (groupids[j] == i+1) {
 	      /* assemble the group cpuset */
 	      hwloc_bitmap_or(group_obj->cpuset, group_obj->cpuset, objs[j]->cpuset);
+	      if (objs[i]->complete_cpuset) {
+		if (!group_obj->complete_cpuset)
+		  group_obj->complete_cpuset = hwloc_bitmap_alloc();
+		hwloc_bitmap_or(group_obj->complete_cpuset, group_obj->complete_cpuset, objs[j]->complete_cpuset);
+	      }
 	      /* if one obj has a nodeset, assemble a group nodeset */
 	      if (objs[j]->nodeset) {
 		if (!group_obj->nodeset)
 		  group_obj->nodeset = hwloc_bitmap_alloc();
 		hwloc_bitmap_or(group_obj->nodeset, group_obj->nodeset, objs[j]->nodeset);
 	      }
+	      if (objs[i]->complete_nodeset) {
+		if (!group_obj->complete_nodeset)
+		  group_obj->complete_nodeset = hwloc_bitmap_alloc();
+		hwloc_bitmap_or(group_obj->complete_nodeset, group_obj->complete_nodeset, objs[j]->complete_nodeset);
+	      }
               groupsizes[i]++;
             }
           hwloc_debug_1arg_bitmap("adding Group object with %u objects and cpuset %s\n",
                                   groupsizes[i], group_obj->cpuset);
           res_obj = hwloc__insert_object_by_cpuset(topology, group_obj,
 						   fromuser ? hwloc_report_user_distance_error : hwloc_report_os_error);
-	  /* res_obj may be different from group_objs if we got groups from XML import before grouping */
+	  /* res_obj may be NULL on failure to insert. */
+	  if (!res_obj)
+	    failed++;
+	  /* or it may be different from groupobjs if we got groups from XML import before grouping */
           groupobjs[i] = res_obj;
       }
 
+      if (failed)
+	/* don't try to group above if we got a NULL group here, just keep this incomplete level */
+	goto inner_free;
+
       /* factorize distances */
       memset(&(groupdistances[0]), 0, sizeof(groupdistances[0]) * nbgroups * nbgroups);
 #undef DISTANCE
@@ -924,7 +976,7 @@ hwloc_group_by_distances(struct hwloc_topology *topology)
 {
   unsigned nbobjs;
   struct hwloc_os_distances_s * osdist;
-  char *env;
+  const char *env;
   float accuracies[5] = { 0.0f, 0.01f, 0.02f, 0.05f, 0.1f };
   unsigned nbaccuracies = 5;
   hwloc_obj_t group_obj;
@@ -1005,12 +1057,22 @@ hwloc_group_by_distances(struct hwloc_topology *topology)
       for(i=0; i<nbobjs; i++) {
 	/* assemble the group cpuset */
 	hwloc_bitmap_or(group_obj->cpuset, group_obj->cpuset, osdist->objs[i]->cpuset);
+	if (osdist->objs[i]->complete_cpuset) {
+	  if (!group_obj->complete_cpuset)
+	    group_obj->complete_cpuset = hwloc_bitmap_alloc();
+	  hwloc_bitmap_or(group_obj->complete_cpuset, group_obj->complete_cpuset, osdist->objs[i]->complete_cpuset);
+	}
 	/* if one obj has a nodeset, assemble a group nodeset */
 	if (osdist->objs[i]->nodeset) {
 	  if (!group_obj->nodeset)
 	    group_obj->nodeset = hwloc_bitmap_alloc();
 	  hwloc_bitmap_or(group_obj->nodeset, group_obj->nodeset, osdist->objs[i]->nodeset);
 	}
+	if (osdist->objs[i]->complete_nodeset) {
+	  if (!group_obj->complete_nodeset)
+	    group_obj->complete_nodeset = hwloc_bitmap_alloc();
+	  hwloc_bitmap_or(group_obj->complete_nodeset, group_obj->complete_nodeset, osdist->objs[i]->complete_nodeset);
+	}
       }
       hwloc_debug_1arg_bitmap("adding Group object (as root of distance matrix with %u objects) with cpuset %s\n",
 			      nbobjs, group_obj->cpuset);
