@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2015 Inria.  All rights reserved.
+ * Copyright © 2009-2017 Inria.  All rights reserved.
  * Copyright © 2009-2010, 2012 Université Bordeaux
  * Copyright © 2009 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -10,6 +10,10 @@
 #include <hwloc-calc.h>
 #include <hwloc.h>
 
+#ifdef HWLOC_LINUX_SYS
+#include <hwloc/linux.h>
+#endif
+
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -40,11 +44,17 @@ void usage(const char *name, FILE *where)
   fprintf(where, "  --get          Retrieve current process binding\n");
   fprintf(where, "  -e --get-last-cpu-location\n"
 		 "                 Retrieve the last processors where the current process ran\n");
+  fprintf(where, "  --nodeset      Display (and parse) cpusets as nodesets\n");
   fprintf(where, "  --pid <pid>    Operate on process <pid>\n");
+#ifdef HWLOC_LINUX_SYS
+  fprintf(where, "  --tid <tid>    Operate on thread <tid>\n");
+#endif
   fprintf(where, "  --taskset      Use taskset-specific format when displaying cpuset strings\n");
   fprintf(where, "Input topology options:\n");
   fprintf(where, "  --restrict <set> Restrict the topology to processors listed in <set>\n");
   fprintf(where, "  --whole-system   Do not consider administration limitations\n");
+  fprintf(where, "  --hbm          Only consider high bandwidth memory nodes\n");
+  fprintf(where, "  --no-hbm       Ignore high-bandwidth memory nodes\n");
   fprintf(where, "Miscellaneous options:\n");
   fprintf(where, "  -f --force     Launch the command even if binding failed\n");
   fprintf(where, "  -q --quiet     Hide non-fatal error messages\n");
@@ -60,21 +70,27 @@ int main(int argc, char *argv[])
   int got_cpubind = 0, got_membind = 0;
   int working_on_cpubind = 1; /* membind if 0 */
   int get_binding = 0;
+  int use_nodeset = 0;
   int get_last_cpu_location = 0;
   unsigned long flags = HWLOC_TOPOLOGY_FLAG_WHOLE_IO|HWLOC_TOPOLOGY_FLAG_ICACHES;
   int force = 0;
   int single = 0;
   int verbose = 0;
+  int only_hbm = -1;
   int logical = 1;
   int taskset = 0;
-  int cpubind_flags = 0;
+  unsigned cpubind_flags = 0;
   hwloc_membind_policy_t membind_policy = HWLOC_MEMBIND_BIND;
-  int membind_flags = 0;
+  int got_mempolicy = 0;
+  unsigned membind_flags = 0;
   int opt;
   int ret;
   int pid_number = -1;
+  int tid_number = -1;
   hwloc_pid_t pid = 0; /* only valid when pid_number > 0, but gcc-4.8 still reports uninitialized warnings */
   char *callname;
+  struct hwloc_calc_location_context_s lcontext;
+  struct hwloc_calc_set_context_s scontext;
 
   cpubind_set = hwloc_bitmap_alloc();
   membind_set = hwloc_bitmap_alloc();
@@ -133,6 +149,17 @@ int main(int argc, char *argv[])
         opt = 1;
         goto next;
       }
+#ifdef HWLOC_LINUX_SYS
+      else if (!strcmp(argv[0], "--tid")) {
+        if (argc < 2) {
+          usage ("hwloc-bind", stderr);
+          exit(EXIT_FAILURE);
+        }
+        tid_number = atoi(argv[1]);
+        opt = 1;
+        goto next;
+      }
+#endif
       else if (!strcmp (argv[0], "--version")) {
           printf("%s %s\n", callname, HWLOC_VERSION);
           exit(EXIT_SUCCESS);
@@ -157,6 +184,10 @@ int main(int argc, char *argv[])
 	get_binding = 1;
 	goto next;
       }
+      else if (!strcmp (argv[0], "--nodeset")) {
+	use_nodeset = 1;
+	goto next;
+      }
       else if (!strcmp (argv[0], "--cpubind")) {
 	  working_on_cpubind = 1;
 	  goto next;
@@ -183,9 +214,18 @@ int main(int argc, char *argv[])
           usage ("hwloc-bind", stderr);
           exit(EXIT_FAILURE);
 	}
+	got_mempolicy = 1;
 	opt = 1;
 	goto next;
       }
+      if (!strcmp(argv[0], "--hbm")) {
+	only_hbm = 1;
+	goto next;
+      }
+      if (!strcmp(argv[0], "--no-hbm")) {
+	only_hbm = 0;
+	goto next;
+      }
       else if (!strcmp (argv[0], "--whole-system")) {
 	flags |= HWLOC_TOPOLOGY_FLAG_WHOLE_SYSTEM;
 	hwloc_topology_destroy(topology);
@@ -207,7 +247,7 @@ int main(int argc, char *argv[])
 	err = hwloc_topology_restrict (topology, restrictset, 0);
 	if (err) {
 	  perror("Restricting the topology");
-	  /* fallthrough */
+	  /* FALLTHRU */
 	}
 	hwloc_bitmap_free(restrictset);
 	argc--;
@@ -220,9 +260,15 @@ int main(int argc, char *argv[])
       return EXIT_FAILURE;
     }
 
-    ret = hwloc_calc_process_arg(topology, depth, argv[0], logical,
-				 working_on_cpubind ? cpubind_set : membind_set,
-				 verbose);
+    lcontext.topology = topology;
+    lcontext.topodepth = depth;
+    lcontext.only_hbm = only_hbm;
+    lcontext.logical = logical;
+    lcontext.verbose = verbose;
+    scontext.nodeset_input = use_nodeset;
+    scontext.nodeset_output = working_on_cpubind ? 0 : 1;
+    scontext.output_set = working_on_cpubind ? cpubind_set : membind_set;
+    ret = hwloc_calc_process_location_as_set(&lcontext, &scontext, argv[0]);
     if (ret < 0) {
       if (verbose > 0)
 	fprintf(stderr, "assuming the command starts at %s\n", argv[0]);
@@ -238,6 +284,11 @@ int main(int argc, char *argv[])
     argv += opt+1;
   }
 
+  if (pid_number > 0 && tid_number > 0) {
+    fprintf(stderr, "cannot operate both on tid and pid\n");
+    return EXIT_FAILURE;
+  }
+
   if (pid_number > 0) {
     pid = hwloc_pid_from_number(pid_number, !(get_binding || get_last_cpu_location));
     /* no need to set_pid()
@@ -263,11 +314,19 @@ int main(int argc, char *argv[])
       if (get_last_cpu_location) {
 	if (pid_number > 0)
 	  err = hwloc_get_proc_last_cpu_location(topology, pid, cpubind_set, 0);
+#ifdef HWLOC_LINUX_SYS
+	else if (tid_number > 0)
+	  err = hwloc_linux_get_tid_last_cpu_location(topology, tid_number, cpubind_set);
+#endif
 	else
 	  err = hwloc_get_last_cpu_location(topology, cpubind_set, 0);
       } else {
 	if (pid_number > 0)
 	  err = hwloc_get_proc_cpubind(topology, pid, cpubind_set, 0);
+#ifdef HWLOC_LINUX_SYS
+	else if (tid_number > 0)
+	  err = hwloc_linux_get_tid_cpubind(topology, tid_number, cpubind_set);
+#endif
 	else
 	  err = hwloc_get_cpubind(topology, cpubind_set, 0);
       }
@@ -275,20 +334,42 @@ int main(int argc, char *argv[])
 	const char *errmsg = strerror(errno);
 	if (pid_number > 0)
 	  fprintf(stderr, "hwloc_get_proc_%s %d failed (errno %d %s)\n", get_last_cpu_location ? "last_cpu_location" : "cpubind", pid_number, errno, errmsg);
+	else if (tid_number > 0)
+	  fprintf(stderr, "hwloc_get_tid_%s %d failed (errno %d %s)\n", get_last_cpu_location ? "last_cpu_location" : "cpubind", tid_number, errno, errmsg);
 	else
 	  fprintf(stderr, "hwloc_get_%s failed (errno %d %s)\n", get_last_cpu_location ? "last_cpu_location" : "cpubind", errno, errmsg);
 	return EXIT_FAILURE;
       }
-      if (taskset)
-	hwloc_bitmap_taskset_asprintf(&s, cpubind_set);
-      else
-	hwloc_bitmap_asprintf(&s, cpubind_set);
-    } else {
+      if (use_nodeset) {
+	hwloc_bitmap_t nset = hwloc_bitmap_alloc();
+	hwloc_cpuset_to_nodeset(topology, cpubind_set, nset);
+	if (taskset)
+	  hwloc_bitmap_taskset_asprintf(&s, nset);
+	else
+	  hwloc_bitmap_asprintf(&s, nset);
+	hwloc_bitmap_free(nset);
+      } else {
+	if (taskset)
+	  hwloc_bitmap_taskset_asprintf(&s, cpubind_set);
+	else
+	  hwloc_bitmap_asprintf(&s, cpubind_set);
+      }
+
+      } else {
       hwloc_membind_policy_t policy;
-      if (pid_number > 0)
-	err = hwloc_get_proc_membind(topology, pid, membind_set, &policy, 0);
-      else
-	err = hwloc_get_membind(topology, membind_set, &policy, 0);
+      if (pid_number > 0) {
+	if (use_nodeset)
+	  err = hwloc_get_proc_membind_nodeset(topology, pid, membind_set, &policy, 0);
+	else
+	  err = hwloc_get_proc_membind(topology, pid, membind_set, &policy, 0);
+      } else if (tid_number > 0) {
+	err = -1; errno = ENOSYS;
+      } else {
+	if (use_nodeset)
+	  err = hwloc_get_membind_nodeset(topology, membind_set, &policy, 0);
+	else
+	  err = hwloc_get_membind(topology, membind_set, &policy, 0);
+      }
       if (err) {
 	const char *errmsg = strerror(errno);
         if (pid_number > 0)
@@ -334,24 +415,29 @@ int main(int argc, char *argv[])
     if (single)
       hwloc_bitmap_singlify(membind_set);
     if (pid_number > 0)
-      ret = hwloc_set_proc_membind(topology, pid, membind_set, membind_policy, membind_flags);
-    else
-      ret = hwloc_set_membind(topology, membind_set, membind_policy, membind_flags);
+      ret = hwloc_set_proc_membind_nodeset(topology, pid, membind_set, membind_policy, membind_flags);
+    else if (tid_number > 0) {
+      ret = -1; errno = ENOSYS;
+    } else
+      ret = hwloc_set_membind_nodeset(topology, membind_set, membind_policy, membind_flags);
     if (ret && verbose >= 0) {
       int bind_errno = errno;
       const char *errmsg = strerror(bind_errno);
       char *s;
       hwloc_bitmap_asprintf(&s, membind_set);
       if (pid_number > 0)
-        fprintf(stderr, "hwloc_set_proc_membind %s (policy %u flags %x) PID %d failed (errno %d %s)\n",
+        fprintf(stderr, "hwloc_set_proc_membind %s (policy %d flags %x) PID %d failed (errno %d %s)\n",
 		s, membind_policy, membind_flags, pid_number, bind_errno, errmsg);
       else
-        fprintf(stderr, "hwloc_set_membind %s (policy %u flags %x) failed (errno %d %s)\n",
+        fprintf(stderr, "hwloc_set_membind %s (policy %d flags %x) failed (errno %d %s)\n",
 		s, membind_policy, membind_flags, bind_errno, errmsg);
       free(s);
     }
     if (ret && !force)
       goto failed_binding;
+  } else {
+    if (got_mempolicy)
+      fprintf(stderr, "--mempolicy ignored unless memory binding is also requested with --membind.\n");
   }
 
   if (got_cpubind) {
@@ -376,6 +462,10 @@ int main(int argc, char *argv[])
       hwloc_bitmap_singlify(cpubind_set);
     if (pid_number > 0)
       ret = hwloc_set_proc_cpubind(topology, pid, cpubind_set, cpubind_flags);
+#ifdef HWLOC_LINUX_SYS
+    else if (tid_number > 0)
+      ret = hwloc_linux_set_tid_cpubind(topology, tid_number, cpubind_set);
+#endif
     else
       ret = hwloc_set_cpubind(topology, cpubind_set, cpubind_flags);
     if (ret && verbose >= 0) {
@@ -386,6 +476,9 @@ int main(int argc, char *argv[])
       if (pid_number > 0)
         fprintf(stderr, "hwloc_set_proc_cpubind %s (flags %x) PID %d failed (errno %d %s)\n",
 		s, cpubind_flags, pid_number, bind_errno, errmsg);
+      else if (tid_number > 0)
+        fprintf(stderr, "hwloc_set_tid_cpubind %s (flags %x) PID %d failed (errno %d %s)\n",
+		s, cpubind_flags, tid_number, bind_errno, errmsg);
       else
         fprintf(stderr, "hwloc_set_cpubind %s (flags %x) failed (errno %d %s)\n",
 		s, cpubind_flags, bind_errno, errmsg);
@@ -400,7 +493,7 @@ int main(int argc, char *argv[])
 
   hwloc_topology_destroy(topology);
 
-  if (pid_number > 0)
+  if (pid_number > 0 || tid_number > 0)
     return EXIT_SUCCESS;
 
   if (0 == argc) {
