@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2016 Inria.  All rights reserved.
+ * Copyright © 2009-2017 Inria.  All rights reserved.
  * Copyright © 2009-2013, 2015 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -90,22 +90,14 @@
 #define BRIDGE_G_COLOR 0xff
 #define BRIDGE_B_COLOR 0xff
 
-struct style {
-  struct stylecolor { int r, g, b; }
-	bg,	/* main box background color */
-	t,	/* main text color */
-	bg2,	/* other box background color */
-	t2;	/* other text color */
-};
-
-unsigned get_textwidth(void *output, struct draw_methods *methods,
-		       const char *text, unsigned length,
-		       unsigned fontsize, unsigned gridsize)
+static unsigned
+get_textwidth(void *output,
+	      const char *text, unsigned length)
 {
   struct lstopo_output *loutput = output;
   if (loutput->methods->textsize) {
     unsigned width;
-    loutput->methods->textsize(output, text, length, fontsize, &width);
+    loutput->methods->textsize(output, text, length, &width);
     return width + gridsize;
   }
   return (length * fontsize * 3) / 4 + gridsize;
@@ -131,7 +123,7 @@ static int prefer_ratio(float ratio1, float ratio2) {
 static void null_declare_color(void *output __hwloc_attribute_unused, int r __hwloc_attribute_unused, int g __hwloc_attribute_unused, int b __hwloc_attribute_unused) { }
 static void null_box(void *output __hwloc_attribute_unused, int r __hwloc_attribute_unused, int g __hwloc_attribute_unused, int b __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x __hwloc_attribute_unused, unsigned width __hwloc_attribute_unused, unsigned y __hwloc_attribute_unused, unsigned height __hwloc_attribute_unused) { }
 static void null_line(void *output __hwloc_attribute_unused, int r __hwloc_attribute_unused, int g __hwloc_attribute_unused, int b __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x1 __hwloc_attribute_unused, unsigned y1_arg __hwloc_attribute_unused, unsigned x2 __hwloc_attribute_unused, unsigned y2 __hwloc_attribute_unused) { }
-static void null_text(void *output __hwloc_attribute_unused, int r __hwloc_attribute_unused, int g __hwloc_attribute_unused, int b __hwloc_attribute_unused, int size __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x __hwloc_attribute_unused, unsigned y __hwloc_attribute_unused, const char *text __hwloc_attribute_unused) { }
+static void null_text(void *output __hwloc_attribute_unused, int r __hwloc_attribute_unused, int g __hwloc_attribute_unused, int b __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x __hwloc_attribute_unused, unsigned y __hwloc_attribute_unused, const char *text __hwloc_attribute_unused) { }
 
 static struct draw_methods null_draw_methods = {
   NULL, /* init */
@@ -185,8 +177,8 @@ static foo_draw get_type_fun(hwloc_obj_type_t type);
   if (obj->type == HWLOC_OBJ_BRIDGE && lstopo_collapse) { \
     for (i = 0; i < numsubobjs; i++) \
       if (subobjs[i]->type == HWLOC_OBJ_PCI_DEVICE) { \
-	const char *collapsestr = hwloc_obj_get_info_by_name(subobjs[i], "lstopoCollapse"); \
-	if (collapsestr && !strcmp(collapsestr, "0")) \
+	const char *_collapsestr = hwloc_obj_get_info_by_name(subobjs[i], "lstopoCollapse"); \
+	if (_collapsestr && !strcmp(_collapsestr, "0")) \
 	  numignoredsubobjs++; \
       } \
   } \
@@ -199,8 +191,8 @@ static foo_draw get_type_fun(hwloc_obj_type_t type);
       if (lstopo_ignore_pus && subobjs[i]->type == HWLOC_OBJ_PU) \
 	continue; \
       if (lstopo_collapse && subobjs[i]->type == HWLOC_OBJ_PCI_DEVICE) { \
-	const char *collapsestr = hwloc_obj_get_info_by_name(subobjs[i], "lstopoCollapse"); \
-	if (collapsestr && !strcmp(collapsestr, "0")) \
+	const char *_collapsestr = hwloc_obj_get_info_by_name(subobjs[i], "lstopoCollapse"); \
+	if (_collapsestr && !strcmp(_collapsestr, "0")) \
 	  continue; \
       }
 
@@ -365,18 +357,9 @@ RECURSE_BEGIN(obj, border) \
 
 /* Pack objects in a grid */
 #define RECURSE_RECT(obj, methods, separator, border) do {\
-  if (obj->arity && obj->children[0]->type == HWLOC_OBJ_NUMANODE && force_orient[obj->type] != LSTOPO_ORIENT_RECT) { \
-    /* Nodes shouldn't be put with an arbitrary geometry, as NUMA distances may not be that way */ \
-    int pvert = prefer_vert(topology, logical, level, output, depth, x, y, separator); \
-    if (pvert) \
-      RECURSE_VERT(level, methods, separator, border); \
-    else \
-      RECURSE_HORIZ(level, methods, separator, border); \
-  } else {\
-    RECURSE_RECT_BEGIN(obj, methods, separator, border) \
-    RECURSE_CALL_FUN(methods); \
-    RECURSE_RECT_END(obj, methods, separator, border); \
-  } \
+  RECURSE_RECT_BEGIN(obj, methods, separator, border) \
+  RECURSE_CALL_FUN(methods); \
+  RECURSE_RECT_END(obj, methods, separator, border); \
 } while (0)
 
 /* Dynamic programming */
@@ -440,23 +423,80 @@ lstopo_obj_snprintf(char *text, size_t textlen, hwloc_obj_t obj, int logical)
     hwloc_obj_type_snprintf(typestr, sizeof(typestr), obj, 0);
   }
 
-  if (idx != (unsigned)-1 && obj->depth != 0
+  if (show_indexes[obj->type]
+      && idx != (unsigned)-1 && obj->depth != 0
       && obj->type != HWLOC_OBJ_PCI_DEVICE
       && (obj->type != HWLOC_OBJ_BRIDGE || obj->attr->bridge.upstream_type == HWLOC_OBJ_BRIDGE_HOST))
     snprintf(indexstr, sizeof(indexstr), "%s%u", indexprefix, idx);
-  attrlen = hwloc_obj_attr_snprintf(attrstr, sizeof(attrstr), obj, " ", 0);
-  /* display the root total_memory if different from the local_memory (already shown) */
-  if (!obj->parent && obj->memory.total_memory > obj->memory.local_memory)
-    snprintf(totmemstr, sizeof(totmemstr), " (%lu%s total)",
-             (unsigned long) hwloc_memory_size_printf_value(obj->memory.total_memory, 0),
-             hwloc_memory_size_printf_unit(obj->memory.total_memory, 0));
+
+  if (show_attrs[obj->type]) {
+    attrlen = hwloc_obj_attr_snprintf(attrstr, sizeof(attrstr), obj, " ", 0);
+    /* display the root total_memory if different from the local_memory (already shown) */
+    if (!obj->parent && obj->memory.total_memory > obj->memory.local_memory)
+      snprintf(totmemstr, sizeof(totmemstr), " (%lu%s total)",
+	       (unsigned long) hwloc_memory_size_printf_value(obj->memory.total_memory, 0),
+	       hwloc_memory_size_printf_unit(obj->memory.total_memory, 0));
+  } else
+    attrlen = 0;
+
   if (attrlen > 0)
     return snprintf(text, textlen, "%s%s (%s)%s", typestr, indexstr, attrstr, totmemstr);
   else
     return snprintf(text, textlen, "%s%s%s", typestr, indexstr, totmemstr);
 }
 
-static struct draw_methods getmax_draw_methods;
+static void
+lstopo_prepare_custom_styles(struct lstopo_output *loutput, hwloc_obj_t obj)
+{
+  struct lstopo_obj_userdata *lud = obj->userdata;
+  struct style *s = &lud->style;
+  hwloc_obj_t child;
+  unsigned forcer, forceg, forceb;
+  const char *stylestr;
+
+  lud->style_set = 0;
+
+  stylestr = hwloc_obj_get_info_by_name(obj, "lstopoStyle");
+  if (stylestr) {
+    while (*stylestr != '\0') {
+      if (sscanf(stylestr, "%02x%02x%02x", &forcer, &forceg, &forceb) == 3
+	  || sscanf(stylestr, "Background=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
+	s->bg.r = forcer & 255;
+	s->bg.g = forceg & 255;
+	s->bg.b = forceb & 255;
+	lud->style_set |= LSTOPO_STYLE_BG;
+	loutput->methods->declare_color(loutput, s->bg.r, s->bg.g, s->bg.b);
+	s->t.r = s->t.g = s->t.b = (s->bg.r + s->bg.g + s->bg.b < 0xff) ? 0xff : 0;
+    } else if (sscanf(stylestr, "Background2=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
+	s->bg2.r = forcer & 255;
+	s->bg2.g = forceg & 255;
+	s->bg2.b = forceb & 255;
+	lud->style_set |= LSTOPO_STYLE_BG2;
+	loutput->methods->declare_color(loutput, s->bg2.r, s->bg2.g, s->bg2.b);
+	s->t2.r = s->t2.g = s->t2.b = (s->bg2.r + s->bg2.g + s->bg2.b < 0xff) ? 0xff : 0;
+      } else if (sscanf(stylestr, "Text=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
+	s->t.r = forcer & 255;
+	s->t.g = forceg & 255;
+	s->t.b = forceb & 255;
+	lud->style_set |= LSTOPO_STYLE_T;
+	loutput->methods->declare_color(loutput, s->t.r, s->t.g, s->t.b);
+      } else if (sscanf(stylestr, "Text2=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
+	s->t2.r = forcer & 255;
+	s->t2.g = forceg & 255;
+	s->t2.b = forceb & 255;
+	lud->style_set |= LSTOPO_STYLE_T2;
+	loutput->methods->declare_color(loutput, s->t2.r, s->t2.g, s->t2.b);
+      }
+      stylestr = strchr(stylestr, ';');
+      if (!stylestr)
+	break;
+      stylestr++;
+    }
+  }
+
+  for(child = obj->first_child; child; child = child->next_sibling)
+    lstopo_prepare_custom_styles(loutput, child);
+}
 
 static void
 lstopo_set_object_color(struct draw_methods *methods,
@@ -465,11 +505,10 @@ lstopo_set_object_color(struct draw_methods *methods,
 						   * Machine status (0=normal, 1=displayed as a root/System) */
 			struct style *s)
 {
-  unsigned forcer, forceg, forceb;
-  const char *style;
+  struct lstopo_obj_userdata *lud = obj->userdata;
 
   /* no need to deal with colors when computing max sizes */
-  if (methods == &getmax_draw_methods)
+  if (methods == &null_draw_methods)
     return;
 
   memset(s, 0, sizeof(*s));
@@ -484,7 +523,7 @@ lstopo_set_object_color(struct draw_methods *methods,
       break;
     }
     assert(arg == 1); /* Machine printed as a System (when root) */
-    /* fallthrough */
+    /* FALLTHRU */
   case HWLOC_OBJ_SYSTEM:
     s->bg.r = SYSTEM_R_COLOR;
     s->bg.g = SYSTEM_G_COLOR;
@@ -580,34 +619,14 @@ lstopo_set_object_color(struct draw_methods *methods,
     assert(0);
   }
 
-  style = hwloc_obj_get_info_by_name(obj, "lstopoStyle");
-  if (style)
-    while (*style != '\0') {
-      if (sscanf(style, "%02x%02x%02x", &forcer, &forceg, &forceb) == 3
-	  || sscanf(style, "Background=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
-	s->bg.r = forcer & 255;
-	s->bg.g = forceg & 255;
-	s->bg.b = forceb & 255;
-	s->t.r = s->t.g = s->t.b = (s->bg.r + s->bg.g + s->bg.b < 0xff) ? 0xff : 0;
-      } else if (sscanf(style, "Background2=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
-	s->bg2.r = forcer & 255;
-	s->bg2.g = forceg & 255;
-	s->bg2.b = forceb & 255;
-	s->t2.r = s->t2.g = s->t2.b = (s->bg2.r + s->bg2.g + s->bg2.b < 0xff) ? 0xff : 0;
-      } else if (sscanf(style, "Text=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
-	s->t.r = forcer & 255;
-	s->t.g = forceg & 255;
-	s->t.b = forceb & 255;
-      } else if (sscanf(style, "Text2=#%02x%02x%02x", &forcer, &forceg, &forceb) == 3) {
-	s->t2.r = forcer & 255;
-	s->t2.g = forceg & 255;
-	s->t2.b = forceb & 255;
-      }
-      style = strchr(style, ';');
-      if (!style)
-	break;
-      style++;
-    }
+  if (lud->style_set & LSTOPO_STYLE_BG)
+    memcpy(&s->bg, &lud->style.bg, sizeof(struct stylecolor));
+  if (lud->style_set & LSTOPO_STYLE_T)
+    memcpy(&s->t, &lud->style.t, sizeof(struct stylecolor));
+  if (lud->style_set & LSTOPO_STYLE_BG2)
+    memcpy(&s->bg2, &lud->style.bg2, sizeof(struct stylecolor));
+  if (lud->style_set & LSTOPO_STYLE_T2)
+    memcpy(&s->t2, &lud->style.t2, sizeof(struct stylecolor));
 }
 
 static void
@@ -634,7 +653,7 @@ pci_device_draw(hwloc_topology_t topology __hwloc_attribute_unused, struct draw_
     } else {
       n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
     }
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   if (collapse > 1) {
@@ -664,7 +683,7 @@ pci_device_draw(hwloc_topology_t topology __hwloc_attribute_unused, struct draw_
   }
 
   if (fontsize)
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
 
   RECURSE_RECT(level, methods, gridsize, gridsize);
 
@@ -674,19 +693,22 @@ pci_device_draw(hwloc_topology_t topology __hwloc_attribute_unused, struct draw_
 static void
 os_device_draw(hwloc_topology_t topology __hwloc_attribute_unused, struct draw_methods *methods, int logical __hwloc_attribute_unused, hwloc_obj_t level, void *output, unsigned depth, unsigned x, unsigned *retwidth, unsigned y, unsigned *retheight)
 {
+  unsigned myheight = 0, totheight;
+  unsigned mywidth = 0, totwidth;
   unsigned textwidth = gridsize;
-  unsigned totheight = gridsize;
-  unsigned totwidth = gridsize;
   struct style style;
   char text[64];
   int n;
-  unsigned nmorelines = 0, i;
+  unsigned nmorelines = 0, j;
   char morelines[3][64];
 
+  DYNA_CHECK();
+
   if (fontsize) {
     const char *coproctype;
 
-    if (HWLOC_OBJ_OSDEV_COPROC == level->attr->osdev.type
+    if (show_attrs[HWLOC_OBJ_OS_DEVICE]
+	&& HWLOC_OBJ_OSDEV_COPROC == level->attr->osdev.type
         && (coproctype = hwloc_obj_get_info_by_name(level, "CoProcType")) != NULL) {
 
       if (!strcmp(coproctype, "CUDA")) {
@@ -754,28 +776,31 @@ os_device_draw(hwloc_topology_t topology __hwloc_attribute_unused, struct draw_m
     }
 
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
-    for(i=0; i<nmorelines; i++) {
-      unsigned nn = (unsigned)strlen(morelines[i]);
-      unsigned ntextwidth = get_textwidth(output, methods, morelines[i], nn, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
+    for(j=0; j<nmorelines; j++) {
+      unsigned nn = (unsigned)strlen(morelines[j]);
+      unsigned ntextwidth = get_textwidth(output, morelines[j], nn);
       if (ntextwidth > textwidth)
 	textwidth = ntextwidth;
     }
-    totheight = gridsize + (fontsize + gridsize)*(nmorelines+1);
-    totwidth = gridsize + textwidth;
+    myheight = (fontsize + gridsize)*(nmorelines+1);
+    mywidth = 0;
   }
 
-  *retwidth = totwidth;
-  *retheight = totheight;
+  RECURSE_RECT(level, &null_draw_methods, gridsize, gridsize);
 
   lstopo_set_object_color(methods, topology, level, 0, &style);
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, *retwidth, y, *retheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
-    for(i=0; i<nmorelines; i++)
-      methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + (i+2)*gridsize + (i+1)*fontsize, morelines[i]);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
+    for(j=0; j<nmorelines; j++)
+      methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + (j+2)*gridsize + (j+1)*fontsize, morelines[j]);
   }
+
+  RECURSE_RECT(level, methods, gridsize, gridsize);
+
+  DYNA_SAVE();
 }
 
 static void
@@ -819,13 +844,13 @@ bridge_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical
           speed = subobjs[i]->attr->pcidev.linkspeed;
         if (subobjs[i]->type == HWLOC_OBJ_BRIDGE && subobjs[i]->attr->bridge.upstream_type == HWLOC_OBJ_BRIDGE_PCI)
           speed = subobjs[i]->attr->bridge.upstream.pci.linkspeed;
-        if (speed != 0.) {
+        if (show_attrs[HWLOC_OBJ_BRIDGE] && speed != 0.) {
           char text[4];
           if (speed >= 10.)
 	    snprintf(text, sizeof(text), "%.0f", subobjs[i]->attr->pcidev.linkspeed);
 	  else
 	    snprintf(text, sizeof(text), "%0.1f", subobjs[i]->attr->pcidev.linkspeed);
-          methods->text(output, style.t2.r, style.t2.g, style.t2.b, fontsize, depth-1, x + 2*gridsize + gridsize, y + totheight, text);
+          methods->text(output, style.t2.r, style.t2.g, style.t2.b, depth-1, x + 2*gridsize + gridsize, y + totheight, text);
         }
       }
     RECURSE_END_VERT(gridsize, 0);
@@ -859,7 +884,7 @@ pu_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical, hw
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
     /* if smaller than other PU, artificially extend/shift it
      * to make PU boxes nicer when vertically stacked.
      */
@@ -883,7 +908,7 @@ pu_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical, hw
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, *retwidth, y, *retheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize + textxoffset, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize + textxoffset, y + gridsize, text);
   }
 
   RECURSE_RECT(level, methods, 0, gridsize);
@@ -907,7 +932,7 @@ cache_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
     textwidth += gridsize; /* artificially extend the minimal inner size because RECURSE_RECT() uses 0 as border when computing totwidth */
   }
 
@@ -917,7 +942,7 @@ cache_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, myheight - gridsize);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   RECURSE_RECT(level, methods, separator, 0);
@@ -939,7 +964,7 @@ core_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   RECURSE_RECT(level, &null_draw_methods, 0, gridsize);
@@ -948,7 +973,7 @@ core_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, totheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   RECURSE_RECT(level, methods, 0, gridsize);
@@ -970,7 +995,7 @@ package_draw(hwloc_topology_t topology, struct draw_methods *methods, int logica
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   RECURSE_RECT(level, &null_draw_methods, gridsize, gridsize);
@@ -979,7 +1004,7 @@ package_draw(hwloc_topology_t topology, struct draw_methods *methods, int logica
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, totheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   RECURSE_RECT(level, methods, gridsize, gridsize);
@@ -1009,7 +1034,7 @@ node_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
     textwidth += 2*gridsize; /* so artificially extend the minimal inner size to include space between space between epoxy and memory box */
   }
 
@@ -1023,7 +1048,7 @@ node_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
   methods->box(output, style.bg2.r, style.bg2.g, style.bg2.b, depth-1, x + gridsize, totwidth - 2 * gridsize, y + gridsize, myheight - gridsize);
 
   if (fontsize) {
-    methods->text(output, style.t2.r, style.t2.g, style.t2.b, fontsize, depth-2, x + 2 * gridsize, y + 2 * gridsize, text);
+    methods->text(output, style.t2.r, style.t2.g, style.t2.b, depth-2, x + 2 * gridsize, y + 2 * gridsize, text);
   }
 
   /* Restart, now really drawing sublevels */
@@ -1047,7 +1072,7 @@ machine_draw(hwloc_topology_t topology, struct draw_methods *methods, int logica
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   RECURSE_RECT(level, &null_draw_methods, gridsize, gridsize);
@@ -1056,7 +1081,7 @@ machine_draw(hwloc_topology_t topology, struct draw_methods *methods, int logica
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, totheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   RECURSE_RECT(level, methods, gridsize, gridsize);
@@ -1120,7 +1145,7 @@ system_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   if (level->arity > 1 && (level->children[0]->type == HWLOC_OBJ_MACHINE || !level->children[0]->cpuset))
@@ -1132,7 +1157,7 @@ system_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, totheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   if (level->arity > 1 && (level->children[0]->type == HWLOC_OBJ_MACHINE || !level->children[0]->cpuset))
@@ -1158,7 +1183,7 @@ group_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
   }
 
   if (level->arity > 1 && (level->children[0]->type == HWLOC_OBJ_MACHINE || !level->children[0]->cpuset))
@@ -1170,7 +1195,7 @@ group_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, totheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   if (level->arity > 1 && (level->children[0]->type == HWLOC_OBJ_MACHINE || !level->children[0]->cpuset))
@@ -1197,7 +1222,7 @@ misc_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
 
   if (fontsize) {
     n = lstopo_obj_snprintf(text, sizeof(text), level, logical);
-    textwidth = get_textwidth(output, methods, text, n, fontsize, gridsize);
+    textwidth = get_textwidth(output, text, n);
     textwidth += gridsize; /* artificially extend the minimal inner size because RECURSE_RECT() uses 0 as border when computing totwidth */
   }
 
@@ -1210,7 +1235,7 @@ misc_draw(hwloc_topology_t topology, struct draw_methods *methods, int logical,
   methods->box(output, style.bg.r, style.bg.g, style.bg.b, depth, x, totwidth, y, boxheight);
 
   if (fontsize) {
-    methods->text(output, style.t.r, style.t.g, style.t.b, fontsize, depth-1, x + gridsize, y + gridsize, text);
+    methods->text(output, style.t.r, style.t.g, style.t.b, depth-1, x + gridsize, y + gridsize, text);
   }
 
   if (level->arity > 1 && (level->children[0]->type == HWLOC_OBJ_MACHINE || !level->children[0]->cpuset))
@@ -1228,6 +1253,7 @@ fig(hwloc_topology_t topology, struct draw_methods *methods, int logical, int le
   time_t t;
   char text[128];
   char hostname[128] = "";
+  const char *forcedhostname = NULL;
   unsigned long hostname_size = sizeof(hostname);
 
   system_draw(topology, methods, logical, level, output, depth, x, &totwidth, y, &totheight);
@@ -1238,17 +1264,21 @@ fig(hwloc_topology_t topology, struct draw_methods *methods, int logical, int le
   if (legend) {
       /* Display the hostname, but only if we're showing *this*
          system */
-    if (hwloc_topology_is_thissystem(topology)) {
+    forcedhostname = hwloc_obj_get_info_by_name(hwloc_get_root_obj(topology), "HostName");
+    if (!forcedhostname && hwloc_topology_is_thissystem(topology)) {
 #if defined(HWLOC_WIN_SYS) && !defined(__CYGWIN__)
       GetComputerName(hostname, &hostname_size);
 #else
       gethostname(hostname, hostname_size);
 #endif
     }
-    if (*hostname) {
-      snprintf(text, sizeof(text), "Host: %s", hostname);
+    if (forcedhostname || *hostname) {
+      if (forcedhostname)
+	snprintf(text, sizeof(text), "Host: %s", forcedhostname);
+      else
+	snprintf(text, sizeof(text), "Host: %s", hostname);
       methods->box(output, 0xff, 0xff, 0xff, depth, 0, totwidth, totheight, gridsize + (gridsize + fontsize)*(3+lstopo_append_legends_nr));
-      methods->text(output, 0, 0, 0, fontsize, depth, gridsize, totheight + gridsize, text);
+      methods->text(output, 0, 0, 0, depth, gridsize, totheight + gridsize, text);
       offset = gridsize + fontsize;
     } else {
       methods->box(output, 0xff, 0xff, 0xff, depth, 0, totwidth, totheight, gridsize + (gridsize + fontsize)*(2+lstopo_append_legends_nr));
@@ -1257,7 +1287,7 @@ fig(hwloc_topology_t topology, struct draw_methods *methods, int logical, int le
 
     /* Display whether we're showing physical or logical IDs */
     snprintf(text, sizeof(text), "Indexes: %s", logical ? "logical" : "physical");
-    methods->text(output, 0, 0, 0, fontsize, depth, gridsize, totheight + gridsize + offset, text);
+    methods->text(output, 0, 0, 0, depth, gridsize, totheight + gridsize + offset, text);
 
     /* Display timestamp */
     t = time(NULL);
@@ -1279,11 +1309,11 @@ fig(hwloc_topology_t topology, struct draw_methods *methods, int logical, int le
       snprintf(text, sizeof(text), "Date: %s", date);
     }
 #endif /* HAVE_STRFTIME */
-    methods->text(output, 0, 0, 0, fontsize, depth, gridsize, totheight + gridsize + offset + fontsize + gridsize, text);
+    methods->text(output, 0, 0, 0, depth, gridsize, totheight + gridsize + offset + fontsize + gridsize, text);
 
     offset = totheight + gridsize + offset + (fontsize + gridsize) * 2;
     for(i=0; i<lstopo_append_legends_nr; i++) {
-      methods->text(output, 0, 0, 0, fontsize, depth, gridsize, offset, lstopo_append_legends[i]);
+      methods->text(output, 0, 0, 0, depth, gridsize, offset, lstopo_append_legends[i]);
       offset += fontsize + gridsize;
     }
   }
@@ -1334,7 +1364,9 @@ output_draw_start(struct lstopo_output *output)
   methods->declare_color(output, SYSTEM_R_COLOR, SYSTEM_G_COLOR, SYSTEM_B_COLOR);
   methods->declare_color(output, MISC_R_COLOR, MISC_G_COLOR, MISC_B_COLOR);
   methods->declare_color(output, PCI_DEVICE_R_COLOR, PCI_DEVICE_G_COLOR, PCI_DEVICE_B_COLOR);
+  methods->declare_color(output, OS_DEVICE_R_COLOR, OS_DEVICE_G_COLOR, OS_DEVICE_B_COLOR);
   methods->declare_color(output, BRIDGE_R_COLOR, BRIDGE_G_COLOR, BRIDGE_B_COLOR);
+  lstopo_prepare_custom_styles(output, hwloc_get_root_obj(output->topology));
 }
 
 static void
@@ -1369,7 +1401,7 @@ output_compute_pu_min_textwidth(struct lstopo_output *output)
   }
 
   n = lstopo_obj_snprintf(text, sizeof(text), lastpu, output->logical);
-  output->min_pu_textwidth = get_textwidth(output, output->methods, text, n, fontsize, gridsize);
+  output->min_pu_textwidth = get_textwidth(output, text, n);
 }
 
 void
